@@ -32,7 +32,7 @@
 #define OPENIMAGEIO_COLOR_H
 
 #include "export.h"
-#include "version.h"
+#include "oiioversion.h"
 
 OIIO_NAMESPACE_ENTER
 {
@@ -71,7 +71,7 @@ public:
     /// If OpenColorIO is not enabled, this will result in an error.
     /// 
     /// Multiple calls to this are potentially expensive.
-    ColorConfig(const char * filename);
+    ColorConfig(string_view filename);
     
     ~ColorConfig();
     
@@ -88,26 +88,104 @@ public:
     int getNumColorSpaces() const;
     
     /// Query the name of the specified ColorSpace.
-    const char * getColorSpaceNameByIndex(int index) const;
+    const char * getColorSpaceNameByIndex (int index) const;
 
     /// Get the name of the color space representing the named role,
     /// or NULL if none could be identified.
-    const char * getColorSpaceNameByRole (const char *role) const;
+    const char * getColorSpaceNameByRole (string_view role) const;
     
+    /// Get the number of Looks defined in this configuration
+    int getNumLooks() const;
+    
+    /// Query the name of the specified Look.
+    const char * getLookNameByIndex (int index) const;
+
     /// Given the specified input and output ColorSpace, construct the
     /// processor.  It is possible that this will return NULL, if the
-    /// inputColorSpace doesnt exist, the outputColorSpace doesn't exist,
-    /// or if the specified transformation is illegal (for example, it may
-    /// require the inversion of a 3D-LUT, etc).   When the user is finished
-    /// with a ColorProcess, deleteColorProcessor should be called.
-    /// ColorProcessor(s) remain valid even if the ColorConfig that created
-    /// them no longer exists.
+    /// inputColorSpace doesnt exist, the outputColorSpace doesn't
+    /// exist, or if the specified transformation is illegal (for
+    /// example, it may require the inversion of a 3D-LUT, etc).  When
+    /// the user is finished with a ColorProcess, deleteColorProcessor
+    /// should be called.  ColorProcessor(s) remain valid even if the
+    /// ColorConfig that created them no longer exists.
     /// 
-    /// Multiple calls to this are potentially expensive.
-    
-    ColorProcessor* createColorProcessor(const char * inputColorSpace,
-                                         const char * outputColorSpace) const;
+    /// Multiple calls to this are potentially expensive, so you should
+    /// call once to create a ColorProcessor to use on an entire image
+    /// (or multiple images), NOT for every scanline or pixel
+    /// separately!
+    ColorProcessor* createColorProcessor (string_view inputColorSpace,
+                                          string_view outputColorSpace) const;
     
+    /// Given the named look(s), input and output color spaces,
+    /// construct a color processor that applies an OCIO look
+    /// transformation.  If inverse==true, construct the inverse
+    /// transformation.  The context_key and context_value can
+    /// optionally be used to establish an extra token/value pair in the
+    /// OCIO context.
+    ///
+    /// It is possible that this will return NULL, if one of the color
+    /// spaces or the look itself doesnt exist or is not allowed.  When
+    /// the user is finished with a ColorProcess, deleteColorProcessor
+    /// should be called.  ColorProcessor(s) remain valid even if the
+    /// ColorConfig that created them no longer exists.
+    /// 
+    /// Multiple calls to this are potentially expensive, so you should
+    /// call once to create a ColorProcessor to use on an entire image
+    /// (or multiple images), NOT for every scanline or pixel
+    /// separately!
+    ColorProcessor* createLookTransform (string_view looks,
+                                         string_view inputColorSpace,
+                                         string_view outputColorSpace,
+                                         bool inverse=false,
+                                         string_view context_key="",
+                                         string_view context_value="") const;
+
+    /// Get the number of displays defined in this configuration
+    int getNumDisplays() const;
+
+    /// Query the name of the specified display.
+    const char * getDisplayNameByIndex (int index) const;
+
+    /// Get the number of views for a given display defined in this configuration
+    int getNumViews (string_view display) const;
+
+    /// Query the name of the specified view for the specified display
+    const char * getViewNameByIndex (string_view display, int index) const;
+
+    /// Query the name of the default display
+    const char * getDefaultDisplayName() const;
+
+    /// Query the name of the default view for the specified display
+    const char * getDefaultViewName (string_view display) const;
+
+    /// Construct a processor to transform from the given color space
+    /// to the color space of the given display and view. You may optionally
+    /// override the looks that are, by default, used with the display/view
+    /// combination. Looks is a potentially comma (or colon) delimited list
+    /// of lookNames, where +/- prefixes are optionally allowed to denote
+    /// forward/inverse transformation (and forward is assumed in the
+    /// absence of either). It is possible to remove all looks from the
+    /// display by passing an empty string. The context_key and context_value
+    /// can optionally be used to establish an extra token/value pair in the
+    /// OCIO context.
+    ///
+    /// It is possible that this will return NULL, if one of the color
+    /// spaces or the display or view doesn't exist or is not allowed.  When
+    /// the user is finished with a ColorProcess, deleteColorProcessor
+    /// should be called.  ColorProcessor(s) remain valid even if the
+    /// ColorConfig that created them no longer exists.
+    ///
+    /// Multiple calls to this are potentially expensive, so you should
+    /// call once to create a ColorProcessor to use on an entire image
+    /// (or multiple images), NOT for every scanline or pixel
+    /// separately!
+    ColorProcessor* createDisplayTransform (string_view display,
+                                            string_view view,
+                                            string_view inputColorSpace,
+                                            string_view looks="",
+                                            string_view context_key="",
+                                            string_view context_value="") const;
+
     /// Delete the specified ColorProcessor
     static void deleteColorProcessor(ColorProcessor * processor);
     
