@@ -55,9 +55,10 @@
 #include <cmath>
 
 #include "export.h"
+#include "oiioversion.h"
+#include "platform.h"
 #include "typedesc.h"   /* Needed for TypeDesc definition */
 #include "paramlist.h"
-#include "version.h"
 
 OIIO_NAMESPACE_ENTER
 {
@@ -127,12 +128,6 @@ public:
     int z_channel;            ///< Index of depth channel, or -1 if not known
     bool deep;                ///< Contains deep data
 
-    // quantize is used for ImageOutput
-    int quant_black;          ///< quantization of black (0.0) level
-    int quant_white;          ///< quantization of white (1.0) level
-    int quant_min;            ///< quantization minimum clamp value
-    int quant_max;            ///< quantization maximum clamp value
-
     /// The above contains all the information that is likely needed for
     /// every image file, and common to all formats.  Rather than bloat
     /// this structure, customize it for new formats, or break back
@@ -144,27 +139,21 @@ public:
     /// the imageio library itself.
     ImageIOParameterList extra_attribs;  ///< Additional attributes
 
-    /// Constructor: given just the data format, set the default quantize
-    /// and set all other channels to something reasonable.
+    /// Constructor: given just the data format, set all other fields to
+    /// something reasonable.
     ImageSpec (TypeDesc format = TypeDesc::UNKNOWN);
 
     /// Constructor for simple 2D scanline image with nothing special.
     /// If fmt is not supplied, default to unsigned 8-bit data.
     ImageSpec (int xres, int yres, int nchans, TypeDesc fmt = TypeDesc::UINT8);
 
-    /// Set the data format, and as a side effect set quantize
-    /// to good defaults for that format
+    /// Set the data format.
     void set_format (TypeDesc fmt);
 
     /// Set the channelnames to reasonable defaults ("R", "G", "B", "A"),
     /// and alpha_channel, based on the number of channels.
     void default_channel_names ();
 
-    /// Given quantization parameters, deduce a TypeDesc that can
-    /// be used without unacceptable loss of significant bits.
-    static TypeDesc format_from_quantize (int quant_black, int quant_white,
-                                          int quant_min, int quant_max);
-
     /// Return the number of bytes for each channel datum, assuming they
     /// are all stored using the data format given by this->format.
     size_t channel_bytes() const { return format.size(); }
@@ -276,72 +265,73 @@ public:
 
     /// Add an optional attribute to the extra attribute list
     ///
-    void attribute (const std::string &name, TypeDesc type, const void *value);
+    void attribute (string_view name, TypeDesc type, const void *value);
 
     /// Add an optional attribute to the extra attribute list.
     ///
-    void attribute (const std::string &name, TypeDesc type, const std::string &value);
+    void attribute (string_view name, TypeDesc type, string_view value);
 
     /// Add an unsigned int attribute
     ///
-    void attribute (const std::string &name, unsigned int value) {
+    void attribute (string_view name, unsigned int value) {
         attribute (name, TypeDesc::UINT, &value);
     }
 
     /// Add an int attribute
     ///
-    void attribute (const std::string &name, int value) {
+    void attribute (string_view name, int value) {
         attribute (name, TypeDesc::INT, &value);
     }
 
     /// Add a float attribute
     ///
-    void attribute (const std::string &name, float value) {
+    void attribute (string_view name, float value) {
         attribute (name, TypeDesc::FLOAT, &value);
     }
 
     /// Add a string attribute
     ///
-    void attribute (const std::string &name, const char *value) {
-        attribute (name, TypeDesc::STRING, &value);
+    void attribute (string_view name, string_view value) {
+        const char *s = value.c_str();
+        attribute (name, TypeDesc::STRING, &s);
     }
 
-    /// Add a string attribute
-    ///
-    void attribute (const std::string &name, const std::string &value) {
-        attribute (name, value.c_str());
-    }
-
-    /// Remove the specified attribute from the list of extra
-    /// attributes. If not found, do nothing.
-    void erase_attribute (const std::string &name,
+    /// Remove the specified attribute from the list of extra_attribs. If
+    /// not found, do nothing.  If searchtype is anything but UNKNOWN,
+    /// restrict matches to only those of the given type. If casesensitive
+    /// is true, the name search will be case-sensitive, otherwise the name
+    /// search will be performed without regard to case (this is the
+    /// default).
+    void erase_attribute (string_view name,
                           TypeDesc searchtype=TypeDesc::UNKNOWN,
                           bool casesensitive=false);
 
-    /// Search for a attribute of the given name in the list of extra
-    /// attributes.
-    ImageIOParameter * find_attribute (const std::string &name,
+    /// Search for an attribute of the given name in the list of
+    /// extra_attribs. If searchtype is anything but UNKNOWN, restrict
+    /// matches to only those of the given type. If casesensitive is true,
+    /// the name search will be case-sensitive, otherwise the name search
+    /// will be performed without regard to case (this is the default).
+    ImageIOParameter * find_attribute (string_view name,
                                        TypeDesc searchtype=TypeDesc::UNKNOWN,
                                        bool casesensitive=false);
-    const ImageIOParameter *find_attribute (const std::string &name,
+    const ImageIOParameter *find_attribute (string_view name,
                                             TypeDesc searchtype=TypeDesc::UNKNOWN,
                                             bool casesensitive=false) const;
 
     /// Simple way to get an integer attribute, with default provided.
     /// Automatically will return an int even if the data is really
     /// unsigned, short, or byte.
-    int get_int_attribute (const std::string &name, int defaultval=0) const;
+    int get_int_attribute (string_view name, int defaultval=0) const;
 
     /// Simple way to get a float attribute, with default provided.
     /// Automatically will return a float even if the data is really
     /// double or half.
-    float get_float_attribute (const std::string &name,
-                               float defaultval=0) const;
+    float get_float_attribute (string_view name, float defaultval=0) const;
 
     /// Simple way to get a string attribute, with default provided.
     ///
-    std::string get_string_attribute (const std::string &name,
-                          const std::string &defaultval = std::string()) const;
+    string_view get_string_attribute (string_view name,
+                           string_view defaultval = string_view()) const;
 
     /// For a given parameter (in this ImageSpec's extra_attribs),
     /// format the value nicely as a string.  If 'human' is true, use
@@ -374,14 +364,15 @@ public:
 
     /// Return teh channelformat of the given channel.
     TypeDesc channelformat (int chan) const {
-        return chan >= (int)channelformats.size() ? format : channelformats[chan];
+        return chan >= 0 && chan < (int)channelformats.size()
+            ? channelformats[chan] : format;
     }
 
     /// Fill in an array of channel formats describing all channels in
     /// the image.  (Note that this differs slightly from the member
     /// data channelformats, which is empty if there are not separate
     /// per-channel formats.)
-    void get_channelformats (std::vector<TypeDesc> &formats) {
+    void get_channelformats (std::vector<TypeDesc> &formats) const {
         formats = channelformats;
         if ((int)formats.size() < nchannels)
             formats.resize (nchannels, format);
@@ -400,7 +391,8 @@ public:
     std::vector<char> data;          // for each sample [z][y][x][c][s]
 
     DeepData () : npixels(0), nchannels(0) { }
-    /// Initialize size and allocate nsamples, pointers
+    /// Initialize size and allocate nsamples, pointers. It is important to
+    /// completely fill in nsamples after init() but before alling alloc().
     void init (int npix, int nchan,
                const TypeDesc *chbegin, const TypeDesc *chend);
     /// After nsamples[] has been filled in, allocate enough scratch space
@@ -410,6 +402,17 @@ public:
     void clear ();
     /// Deallocate all space in the vectors
     void free ();
+    /// Retrieve the pointer to the first sample of the given pixel and
+    /// channel. Return NULL if there are no samples for that pixel.
+    void *channel_ptr (int pixel, int channel) const;
+    /// Retrieve deep sample value within a pixel, cast to a float.
+    float deep_value (int pixel, int channel, int sample) const;
+    /// Retrieve deep sample value within a pixel, as an untigned int.
+    uint32_t deep_value_uint (int pixel, int channel, int sample) const;
+    /// Set deep sample value within a pixel, as a float.
+    void set_deep_value (int pixel, int channel, int sample, float value);
+    /// Set deep sample value within a pixel, as a uint32.
+    void set_deep_value_uint (int pixel, int channel, int sample, uint32_t value);
 };
 
 
@@ -456,6 +459,24 @@ public:
     ///
     virtual const char *format_name (void) const = 0;
 
+    /// Given the name of a 'feature', return whether this ImageInput
+    /// supports input of images with the given properties.
+    /// Feature names that ImageIO plugins are expected to recognize
+    /// include:
+    ///    "arbitrary_metadata" Does this format allow metadata with
+    ///                        arbitrary names and types?
+    ///    "exif"           Can this format store Exif camera data?
+    ///    "iptc"           Can this format store IPTC data?
+    ///    "procedural"     Can this format create images without reading
+    ///                        from a disk file?
+    ///
+    /// Note that main advantage of this approach, versus having
+    /// separate individual supports_foo() methods, is that this allows
+    /// future expansion of the set of possible queries without changing
+    /// the API, adding new entry points, or breaking linkage
+    /// compatibility.
+    virtual bool supports (const std::string & /*feature*/) const { return false; }
+
     /// Return true if the named file is file of the type for this
     /// ImageInput.  The implementation will try to determine this as
     /// efficiently as possible, in most cases much less expensively
@@ -486,19 +507,6 @@ public:
     /// a call to seek_subimage().
     const ImageSpec &spec (void) const { return m_spec; }
 
-    /// Given the name of a 'feature', return whether this ImageInput
-    /// supports input of images with the given properties.
-    /// Feature names that ImageIO plugins are expected to recognize
-    /// include:
-    ///    none at this time
-    ///
-    /// Note that main advantage of this approach, versus having
-    /// separate individual supports_foo() methods, is that this allows
-    /// future expansion of the set of possible queries without changing
-    /// the API, adding new entry points, or breaking linkage
-    /// compatibility.
-    virtual bool supports (const std::string & /*feature*/) const { return false; }
-
     /// Close an image that we are totally done with.
     ///
     virtual bool close () = 0;
@@ -593,10 +601,10 @@ public:
 
     /// Read the tile whose upper-left origin is (x,y,z) into data,
     /// converting if necessary from the native data format of the file
-    /// into the 'format' specified.  (z==0 for non-volume images.)  The
-    /// stride values give the data spacing of adjacent pixels,
-    /// scanlines, and volumetric slices (measured in bytes).  Strides
-    /// set to AutoStride imply 'contiguous' data, i.e.,
+    /// into the 'format' specified. (z==0 for non-volume images.) The
+    /// stride values give the data spacing of adjacent pixels, scanlines,
+    /// and volumetric slices (measured in bytes). Strides set to AutoStride
+    /// imply 'contiguous' data in the shape of a full tile, i.e.,
     ///     xstride == spec.nchannels*format.size()
     ///     ystride == xstride*spec.tile_width
     ///     zstride == ystride*spec.tile_height
@@ -634,6 +642,13 @@ public:
     /// boundaries, with the exception that it may also be the end of
     /// the image data if the image resolution is not a whole multiple
     /// of the tile size.
+    /// The stride values give the data spacing of adjacent pixels,
+    /// scanlines, and volumetric slices (measured in bytes). Strides set to
+    /// AutoStride imply 'contiguous' data in the shape of the [begin,end)
+    /// region, i.e.,
+    ///     xstride == spec.nchannels*format.size()
+    ///     ystride == xstride * (xend-xbegin)
+    ///     zstride == ystride * (yend-ybegin)
     virtual bool read_tiles (int xbegin, int xend, int ybegin, int yend,
                              int zbegin, int zend, TypeDesc format,
                              void *data, stride_t xstride=AutoStride,
@@ -774,6 +789,11 @@ public:
         return e;
     }
 
+    /// An ImageInput::Creator is a function that creates and returns an
+    /// ImageInput.  Once invoked, the resulting ImageInput is owned by
+    /// the caller, who is responsible for deleting it when done with it.
+    typedef ImageInput* (*Creator)();
+
 protected:
     /// Error reporting for the plugin implementation: call this with
     /// printf-like arguments.  Note however that this is fully typesafe!
@@ -837,6 +857,9 @@ public:
     ///    "mipmap"         Does this format support multiple resolutions
     ///                       for an image/subimage?
     ///    "volumes"        Does this format support "3D" pixel arrays?
+    ///    "alpha"          Can this format support an alpha channel?
+    ///    "nchannels"      Can this format support arbitrary number of
+    ///                        channels (beyond RGBA)?
     ///    "rewrite"        May the same scanline or tile be sent more than
     ///                       once?  (Generally, this will be true for
     ///                       plugins that implement interactive display.)
@@ -852,13 +875,17 @@ public:
     ///    "negativeorigin" Does the format support negative x,y,z
     ///                        and full_{x,y,z} origin values?
     ///    "deepdata"       Deep (multi-sample per pixel) data
+    ///    "arbitrary_metadata" Does this format allow metadata with
+    ///                        arbitrary names and types?
+    ///    "exif"           Can this format store Exif camera data?
+    ///    "iptc"           Can this format store IPTC data?
     ///
     /// Note that main advantage of this approach, versus having
     /// separate individual supports_foo() methods, is that this allows
     /// future expansion of the set of possible queries without changing
     /// the API, adding new entry points, or breaking linkage
     /// compatibility.
-    virtual bool supports (const std::string & /*feature*/) const { return false; }
+    virtual bool supports (const std::string &feature) const { return false; }
 
     enum OpenMode { Create, AppendSubimage, AppendMIPLevel };
 
@@ -935,7 +962,8 @@ public:
     /// ignored for 2D non-volume images.)  The three stride values give
     /// the distance (in bytes) between successive pixels, scanlines,
     /// and volumetric slices, respectively.  Strides set to AutoStride
-    /// imply 'contiguous' data, i.e.,
+    /// imply 'contiguous' data in the shape of a full tile, i.e.,
+
     ///     xstride == spec.nchannels*format.size()
     ///     ystride == xstride*spec.tile_width
     ///     zstride == ystride*spec.tile_height
@@ -953,13 +981,20 @@ public:
                              stride_t zstride=AutoStride);
 
     /// Write the block of multiple tiles that include all pixels in
-    /// [xbegin,xend) X [ybegin,yend) X [zbegin,zend).  This is
-    /// analogous to write_tile except that it may be used to write more
-    /// than one tile at a time (which, for some formats, may be able to
-    /// be done much more efficiently or in parallel).
-    /// The begin/end pairs must correctly delineate tile boundaries,
-    /// with the exception that it may also be the end of the image data
-    /// if the image resolution is not a whole multiple of the tile size.
+    /// [xbegin,xend) X [ybegin,yend) X [zbegin,zend).  This is analogous to
+    /// write_tile except that it may be used to write more than one tile at
+    /// a time (which, for some formats, may be able to be done much more
+    /// efficiently or in parallel).
+    /// The begin/end pairs must correctly delineate tile boundaries, with
+    /// the exception that it may also be the end of the image data if the
+    /// image resolution is not a whole multiple of the tile size.
+    /// The stride values give the data spacing of adjacent pixels,
+    /// scanlines, and volumetric slices (measured in bytes). Strides set to
+    /// AutoStride imply 'contiguous' data in the shape of the [begin,end)
+    /// region, i.e.,
+    ///     xstride == spec.nchannels*format.size()
+    ///     ystride == xstride * (xend-xbegin)
+    ///     zstride == ystride * (yend-ybegin)
     virtual bool write_tiles (int xbegin, int xend, int ybegin, int yend,
                               int zbegin, int zend, TypeDesc format,
                               const void *data, stride_t xstride=AutoStride,
@@ -968,13 +1003,13 @@ public:
 
     /// Write a rectangle of pixels given by the range
     ///   [xbegin,xend) X [ybegin,yend) X [zbegin,zend)
-    /// The three stride values give the distance (in bytes) between
-    /// successive pixels, scanlines, and volumetric slices,
-    /// respectively.  Strides set to AutoStride imply 'contiguous'
-    /// data, i.e.,
+    /// The stride values give the data spacing of adjacent pixels,
+    /// scanlines, and volumetric slices (measured in bytes). Strides set to
+    /// AutoStride imply 'contiguous' data in the shape of the [begin,end)
+    /// region, i.e.,
     ///     xstride == spec.nchannels*format.size()
-    ///     ystride == xstride * (xmax-xmin+1)
-    ///     zstride == ystride * (ymax-ymin+1)
+    ///     ystride == xstride * (xend-xbegin)
+    ///     zstride == ystride * (yend-ybegin)
     /// The data are automatically converted from 'format' to the actual
     /// output format (as specified to open()) by this method.  If
     /// format is TypeDesc::UNKNOWN, it will just copy pixels assuming
@@ -1064,6 +1099,11 @@ public:
         return e;
     }
 
+    /// An ImageOutput::Creator is a function that creates and returns an
+    /// ImageOutput.  Once invoked, the resulting ImageOutput is owned by
+    /// the caller, who is responsible for deleting it when done with it.
+    typedef ImageOutput* (*Creator)();
+
 protected:
     /// Error reporting for the plugin implementation: call this with
     /// printf-like arguments.  Note however that this is fully typesafe!
@@ -1073,25 +1113,41 @@ protected:
 
     /// Helper routines used by write_* implementations: convert data (in
     /// the given format and stride) to the "native" format of the file
-    /// (described by the 'spec' member variable), in contiguous order.
-    /// This requires a scratch space to be passed in so that there are
-    /// no memory leaks.  Returns a pointer to the native data, which may
-    /// be the original data if it was already in native format and
-    /// contiguous, or it may point to the scratch space if it needed to
-    /// make a copy or do conversions.
+    /// (described by the 'spec' member variable), in contiguous order. This
+    /// requires a scratch space to be passed in so that there are no memory
+    /// leaks.  Returns a pointer to the native data, which may be the
+    /// original data if it was already in native format and contiguous, or
+    /// it may point to the scratch space if it needed to make a copy or do
+    /// conversions. For float->uint8 conversions only, if dither is
+    /// nonzero, random dither will be added to reduce quantization banding
+    /// artifacts; in this case, the specific nonzero dither value is used
+    /// as a seed for the hash function that produces the per-pixel dither
+    /// amounts, and the optional [xyz]origin parameters help it to align
+    /// the pixels to the right position in the dither pattern.
     const void *to_native_scanline (TypeDesc format,
                                     const void *data, stride_t xstride,
-                                    std::vector<unsigned char> &scratch);
+                                    std::vector<unsigned char> &scratch,
+                                    unsigned int dither=0,
+                                    int yorigin=0, int zorigin=0);
     const void *to_native_tile (TypeDesc format, const void *data,
                                 stride_t xstride, stride_t ystride,
                                 stride_t zstride,
-                                std::vector<unsigned char> &scratch);
+                                std::vector<unsigned char> &scratch,
+                                unsigned int dither=0,
+                                int xorigin=0, int yorigin=0, int zorigin=0);
     const void *to_native_rectangle (int xbegin, int xend, int ybegin, int yend,
                                      int zbegin, int zend,
                                      TypeDesc format, const void *data,
                                      stride_t xstride, stride_t ystride,
                                      stride_t zstride,
-                                     std::vector<unsigned char> &scratch);
+                                     std::vector<unsigned char> &scratch,
+                                     unsigned int dither=0,
+                                     int xorigin=0, int yorigin=0, int zorigin=0);
+    /// Helper function to copy a tile of data into an image-sized buffer.
+    bool copy_tile_to_image_buffer (int x, int y, int z, TypeDesc format,
+                                    const void *data, stride_t xstride,
+                                    stride_t ystride, stride_t zstride,
+                                    void *image_buffer);
 
 protected:
     ImageSpec m_spec;           ///< format spec of the currently open image
@@ -1137,19 +1193,15 @@ OIIO_API std::string geterror ();
 ///             are presumed to be used for that format.  Semicolons
 ///             separate the lists for formats.  For example,
 ///                "tiff:tif;jpeg:jpg,jpeg;openexr:exr"
-OIIO_API bool attribute (const std::string &name, TypeDesc type,
-                          const void *val);
+OIIO_API bool attribute (string_view name, TypeDesc type, const void *val);
 // Shortcuts for common types
-inline bool attribute (const std::string &name, int val) {
+inline bool attribute (string_view name, int val) {
     return attribute (name, TypeDesc::TypeInt, &val);
 }
-inline bool attribute (const std::string &name, float val) {
+inline bool attribute (string_view name, float val) {
     return attribute (name, TypeDesc::TypeFloat, &val);
 }
-inline bool attribute (const std::string &name, const char *val) {
-    return attribute (name, TypeDesc::TypeString, &val);
-}
-inline bool attribute (const std::string &name, const std::string &val) {
+inline bool attribute (string_view name, string_view val) {
     const char *s = val.c_str();
     return attribute (name, TypeDesc::TypeString, &s);
 }
@@ -1159,41 +1211,45 @@ inline bool attribute (const std::string &name, const std::string &val) {
 /// otherwise return false and do not modify the contents of *val.  It
 /// is up to the caller to ensure that val points to the right kind and
 /// size of storage for the given type.
-OIIO_API bool getattribute (const std::string &name, TypeDesc type,
+OIIO_API bool getattribute (string_view name, TypeDesc type,
                              void *val);
 // Shortcuts for common types
-inline bool getattribute (const std::string &name, int &val) {
+inline bool getattribute (string_view name, int &val) {
     return getattribute (name, TypeDesc::TypeInt, &val);
 }
-inline bool getattribute (const std::string &name, float &val) {
+inline bool getattribute (string_view name, float &val) {
     return getattribute (name, TypeDesc::TypeFloat, &val);
 }
-inline bool getattribute (const std::string &name, char **val) {
+inline bool getattribute (string_view name, char **val) {
     return getattribute (name, TypeDesc::TypeString, val);
 }
-inline bool getattribute (const std::string &name, std::string &val) {
-    const char *s = NULL;
+inline bool getattribute (string_view name, std::string &val) {
+    ustring s;
     bool ok = getattribute (name, TypeDesc::TypeString, &s);
     if (ok)
-        val = s ? s : "";
+        val = s.string();
     return ok;
 }
 
 
-/// Helper routine: quantize a value to an integer given the
-/// quantization parameters.
-OIIO_API int quantize (float value, int quant_black, int quant_white,
-                        int quant_min, int quant_max);
+/// Register the input and output 'create' routines and list of file
+/// extensions for a particular format.
+OIIO_API void declare_imageio_format (const std::string &format_name,
+                                      ImageInput::Creator input_creator,
+                                      const char **input_extensions,
+                                      ImageOutput::Creator output_creator,
+                                      const char **output_extensions);
+
 
 /// Helper function: convert contiguous arbitrary data between two
 /// arbitrary types (specified by TypeDesc's)
 /// Return true if ok, false if it didn't know how to do the
-/// conversion.  If dst_type is UNKNWON, it will be assumed to be the
+/// conversion.  If dst_type is UNKNOWN, it will be assumed to be the
 /// same as src_type.
 OIIO_API bool convert_types (TypeDesc src_type, const void *src,
                               TypeDesc dst_type, void *dst, int n);
 
-/// DEPRECATED -- for some reason we had a convert_types that took
+/// DEPRECATED(1.4) -- for some reason we had a convert_types that took
 /// alpha_channel and z_channel parameters, but never did anything
 /// with them.
 OIIO_API bool convert_types (TypeDesc src_type, const void *src,
@@ -1210,14 +1266,46 @@ OIIO_API bool convert_types (TypeDesc src_type, const void *src,
 /// data.  Return true if ok, false if it didn't know how to do the
 /// conversion.
 OIIO_API bool convert_image (int nchannels, int width, int height, int depth,
-                              const void *src, TypeDesc src_type,
-                              stride_t src_xstride, stride_t src_ystride,
-                              stride_t src_zstride,
-                              void *dst, TypeDesc dst_type,
-                              stride_t dst_xstride, stride_t dst_ystride,
-                              stride_t dst_zstride,
-                              int alpha_channel = -1, int z_channel = -1);
+                             const void *src, TypeDesc src_type,
+                             stride_t src_xstride, stride_t src_ystride,
+                             stride_t src_zstride,
+                             void *dst, TypeDesc dst_type,
+                             stride_t dst_xstride, stride_t dst_ystride,
+                             stride_t dst_zstride,
+                             int alpha_channel = -1, int z_channel = -1);
 
+/// A version of convert_image that will break up big jobs into multiple
+/// threads.
+OIIO_API bool parallel_convert_image (
+               int nchannels, int width, int height, int depth,
+               const void *src, TypeDesc src_type,
+               stride_t src_xstride, stride_t src_ystride,
+               stride_t src_zstride,
+               void *dst, TypeDesc dst_type,
+               stride_t dst_xstride, stride_t dst_ystride,
+               stride_t dst_zstride,
+               int alpha_channel=-1, int z_channel=-1, int nthreads=0);
+
+/// Add random [-theramplitude,ditheramplitude] dither to the color channels
+/// of the image.  Dither will not be added to the alpha or z channel.  The
+/// image origin and dither seed values allow a reproducible (or variable)
+/// dither pattern.
+OIIO_API void add_dither (int nchannels, int width, int height, int depth,
+                          float *data,
+                          stride_t xstride, stride_t ystride, stride_t zstride,
+                          float ditheramplitude,
+                          int alpha_channel = -1, int z_channel = -1,
+                          unsigned int ditherseed = 1,
+                          int chorigin=0, int xorigin=0,
+                          int yorigin=0, int zorigin=0);
+
+/// Convert unassociated to associated alpha by premultiplying all color
+/// (non-alpha, non-z) channels by alpha.
+OIIO_API void premult (int nchannels, int width, int height, int depth,
+                       int chbegin, int chend,
+                       TypeDesc datatype, void *data, stride_t xstride,
+                       stride_t ystride, stride_t zstride,
+                       int alpha_channel = -1, int z_channel = -1);
 
 /// Helper routine for data conversion: Copy an image of nchannels x
 /// width x height x depth from src to dst.  The src and dst may have
@@ -1228,11 +1316,11 @@ OIIO_API bool convert_image (int nchannels, int width, int height, int depth,
 /// auto-computed assuming contiguous data.  Return true if ok, false if
 /// it didn't know how to do the conversion.
 OIIO_API bool copy_image (int nchannels, int width, int height, int depth,
-                           const void *src, stride_t pixelsize,
-                           stride_t src_xstride, stride_t src_ystride,
-                           stride_t src_zstride,
-                           void *dst, stride_t dst_xstride,
-                           stride_t dst_ystride, stride_t dst_zstride);
+                          const void *src, stride_t pixelsize,
+                          stride_t src_xstride, stride_t src_ystride,
+                          stride_t src_zstride,
+                          void *dst, stride_t dst_xstride,
+                          stride_t dst_ystride, stride_t dst_zstride);
 
 /// Decode a raw Exif data block and save all the metadata in an
 /// ImageSpec.  Return true if all is ok, false if the exif block was
@@ -1244,6 +1332,12 @@ OIIO_API bool decode_exif (const void *exif, int length, ImageSpec &spec);
 /// data as a big blob to the char vector.
 OIIO_API void encode_exif (const ImageSpec &spec, std::vector<char> &blob);
 
+/// Helper: For the given OIIO metadata attribute name, look up the Exif tag
+/// ID, TIFFDataType (expressed as an int), and count. Return true and fill
+/// in the fields if found, return false if not found.
+OIIO_API bool exif_tag_lookup (string_view name, int &tag,
+                               int &tifftype, int &count);
+
 /// Add metadata to spec based on raw IPTC (International Press
 /// Telecommunications Council) metadata in the form of an IIM
 /// (Information Interchange Model).  Return true if all is ok, false if
@@ -1277,6 +1371,21 @@ OIIO_API bool decode_xmp (const std::string &xml, ImageSpec &spec);
 /// be part of ordinary TIFF or exif tags.
 OIIO_API std::string encode_xmp (const ImageSpec &spec, bool minimal=false);
 
+// All the wrap_foo functions implement a wrap mode, wherein coord is
+// altered to be origin <= coord < origin+width.  The return value
+// indicates if the resulting wrapped value is valid (example, for
+// wrap_black, values outside the region are invalid and do not modify
+// the coord parameter).
+OIIO_API bool wrap_black (int &coord, int origin, int width);
+OIIO_API bool wrap_clamp (int &coord, int origin, int width);
+OIIO_API bool wrap_periodic (int &coord, int origin, int width);
+OIIO_API bool wrap_periodic_pow2 (int &coord, int origin, int width);
+OIIO_API bool wrap_mirror (int &coord, int origin, int width);
+
+// Typedef for the function signature of a wrap implementation.
+typedef bool (*wrap_impl) (int &coord, int origin, int width);
+
+
 // to force correct linkage on some systems
 OIIO_API void _ImageIO_force_link ();
 
