@@ -39,8 +39,8 @@
 #ifndef OPENIMAGEIO_THREAD_H
 #define OPENIMAGEIO_THREAD_H
 
-#include "version.h"
-#include "sysutil.h"
+#include "oiioversion.h"
+#include "platform.h"
 
 
 // defining NOMINMAX to prevent problems with std::min/std::max
@@ -70,34 +70,16 @@
 #pragma GCC diagnostic error "-Wunused-variable"
 #endif
 
-#ifndef USE_TBB
-#  define USE_TBB 0
-#endif
-
-// Include files we need for atomic counters.
-// Some day, we hope this is all replaced by use of std::atomic<>.
-#if USE_TBB
-#  include <tbb/atomic.h>
-#  include <tbb/spin_mutex.h>
-#  define USE_TBB_ATOMIC 1
-#  define USE_TBB_SPINLOCK 1
-#else
-#  define USE_TBB_ATOMIC 0
-#  define USE_TBB_SPINLOCK 0
-#endif
-
-
-#if defined(_MSC_VER) && !USE_TBB
+#if defined(_MSC_VER)
 #  include <windows.h>
 #  include <winbase.h>
 #  pragma intrinsic (_InterlockedExchangeAdd)
 #  pragma intrinsic (_InterlockedCompareExchange)
 #  pragma intrinsic (_InterlockedCompareExchange64)
-#  pragma intrinsic (_ReadWriteBarrier)
 #  if defined(_WIN64)
 #    pragma intrinsic(_InterlockedExchangeAdd64)
 #  endif
-// InterlockedExchangeAdd64 is not available for XP
+// InterlockedExchangeAdd64 & InterlockedExchange64 are not available for XP
 #  if defined(_WIN32_WINNT) && _WIN32_WINNT <= 0x0501
 inline long long
 InterlockedExchangeAdd64 (volatile long long *Addend, long long Value)
@@ -108,13 +90,38 @@ InterlockedExchangeAdd64 (volatile long long *Addend, long long Value)
     } while (_InterlockedCompareExchange64(Addend, Old + Value, Old) != Old);
     return Old;
 }
+
+inline long long
+InterlockedExchange64 (volatile long long *Target, long long Value)
+{
+    long long Old;
+    do {
+        Old = *Target;
+    } while (_InterlockedCompareExchange64(Target, Value, Old) != Old);
+    return Old;
+}
 #  endif
 #endif
 
 #if defined(__GNUC__) && (defined(_GLIBCXX_ATOMIC_BUILTINS) || (__GNUC__ * 100 + __GNUC_MINOR__ >= 401))
 #define USE_GCC_ATOMICS
+#  if !defined(__clang__) && (__GNUC__ * 100 + __GNUC_MINOR__ >= 408)
+#    define OIIO_USE_GCC_NEW_ATOMICS
+#  endif
 #endif
 
+
+// OIIO_THREAD_ALLOW_DCLP, if set to 0, prevents us from using a dodgy
+// "double checked lock pattern" (DCLP).  We are very careful to construct
+// it safely and correctly, and these uses improve thread performance for
+// us.  But it confuses Thread Sanitizer, so this switch allows you to turn
+// it off. Also set to 0 if you don't believe that we are correct in
+// allowing this construct on all platforms.
+#ifndef OIIO_THREAD_ALLOW_DCLP
+#define OIIO_THREAD_ALLOW_DCLP 1
+#endif
+
+
 OIIO_NAMESPACE_ENTER
 {
 
@@ -222,11 +229,10 @@ atomic_exchange_and_add (volatile int *at, int x)
 {
 #ifdef NOTHREADS
     int r = *at;  *at += x;  return r;
+#elif defined(OIIO_USE_GCC_NEW_ATOMICS)
+    return __atomic_fetch_add (at, x, __ATOMIC_SEQ_CST);
 #elif defined(USE_GCC_ATOMICS)
     return __sync_fetch_and_add ((int *)at, x);
-#elif USE_TBB
-    atomic<int> *a = (atomic<int> *)at;
-    return a->fetch_and_add (x);
 #elif defined(_MSC_VER)
     // Windows
     return _InterlockedExchangeAdd ((volatile LONG *)at, x);
@@ -242,11 +248,10 @@ atomic_exchange_and_add (volatile long long *at, long long x)
 {
 #ifdef NOTHREADS
     long long r = *at;  *at += x;  return r;
+#elif defined(OIIO_USE_GCC_NEW_ATOMICS)
+    return __atomic_fetch_add (at, x, __ATOMIC_SEQ_CST);
 #elif defined(USE_GCC_ATOMICS)
     return __sync_fetch_and_add (at, x);
-#elif USE_TBB
-    atomic<long long> *a = (atomic<long long> *)at;
-    return a->fetch_and_add (x);
 #elif defined(_MSC_VER)
     // Windows
 #  if defined(_WIN64)
@@ -276,11 +281,11 @@ atomic_compare_and_exchange (volatile int *at, int compareval, int newval)
     } else {
         return false;
     }
+#elif defined(OIIO_USE_GCC_NEW_ATOMICS)
+    return __atomic_compare_exchange_n (at, &compareval, newval, false,
+                                        __ATOMIC_SEQ_CST, __ATOMIC_SEQ_CST);
 #elif defined(USE_GCC_ATOMICS)
     return __sync_bool_compare_and_swap (at, compareval, newval);
-#elif USE_TBB
-    atomic<int> *a = (atomic<int> *)at;
-    return a->compare_and_swap (newval, compareval) == newval;
 #elif defined(_MSC_VER)
     return (_InterlockedCompareExchange ((volatile LONG *)at, newval, compareval) == compareval);
 #else
@@ -299,11 +304,11 @@ atomic_compare_and_exchange (volatile long long *at, long long compareval, long
     } else {
         return false;
     }
+#elif defined(OIIO_USE_GCC_NEW_ATOMICS)
+    return __atomic_compare_exchange_n (at, &compareval, newval, false,
+                                        __ATOMIC_SEQ_CST, __ATOMIC_SEQ_CST);
 #elif defined(USE_GCC_ATOMICS)
     return __sync_bool_compare_and_swap (at, compareval, newval);
-#elif USE_TBB
-    atomic<long long> *a = (atomic<long long> *)at;
-    return a->compare_and_swap (newval, compareval) == newval;
 #elif defined(_MSC_VER)
     return (_InterlockedCompareExchange64 ((volatile LONGLONG *)at, newval, compareval) == compareval);
 #else
@@ -313,6 +318,64 @@ atomic_compare_and_exchange (volatile long long *at, long long compareval, long
 
 
 
+/// Atomic version of:  r = *at, *at = x, return r
+/// For each of several architectures.
+inline int
+atomic_exchange (volatile int *at, int x)
+{
+#ifdef NOTHREADS
+    int r = *at;  *at = x;  return r;
+#elif defined(OIIO_USE_GCC_NEW_ATOMICS)
+    return __atomic_exchange_n (at, x, __ATOMIC_SEQ_CST);
+#elif defined(USE_GCC_ATOMICS)
+    // No __sync version of atomic exchange! Do it the hard way:
+    while (1) {
+        int old = *at;
+        if (atomic_compare_and_exchange (at, old, x))
+            return old;
+    }
+    return 0; // can never happen
+#elif defined(_MSC_VER)
+    // Windows
+    return _InterlockedExchange ((volatile LONG *)at, x);
+#else
+#   error No atomics on this platform.
+#endif
+}
+
+
+
+inline long long
+atomic_exchange (volatile long long *at, long long x)
+{
+#ifdef NOTHREADS
+    long long r = *at;  *at = x;  return r;
+#elif defined(OIIO_USE_GCC_NEW_ATOMICS)
+    return __atomic_exchange_n (at, x, __ATOMIC_SEQ_CST);
+#elif defined(USE_GCC_ATOMICS)
+    // No __sync version of atomic exchange! Do it the hard way:
+    while (1) {
+        long long old = *at;
+        if (atomic_compare_and_exchange (at, old, x))
+            return old;
+    }
+    return 0; // can never happen
+#elif defined(_MSC_VER)
+    // Windows
+#  if defined(_WIN64)
+    return _InterlockedExchange64 ((volatile LONGLONG *)at, x);
+#  else
+    return InterlockedExchange64 ((volatile LONGLONG *)at, x);
+#  endif
+#else
+#   error No atomics on this platform.
+#endif
+}
+
+
+
+
+
 /// Yield the processor for the rest of the timeslice.
 ///
 inline void
@@ -341,9 +404,6 @@ pause (int delay)
     for (int i = 0; i < delay; ++i)
         __asm__ __volatile__("NOP;");
 
-#elif USE_TBB
-    __TBB_Pause(delay);
-
 #elif defined(_MSC_VER)
     for (int i = 0; i < delay; ++i) {
 #if defined (_WIN64)
@@ -381,12 +441,6 @@ private:
 
 
 
-#if USE_TBB_ATOMIC
-using tbb::atomic;
-#else
-// If we're not using TBB's atomic, we need to define our own atomic<>.
-
-
 /// Atomic integer.  Increment, decrement, add, and subtract in a
 /// totally thread-safe manner.
 template<class T>
@@ -467,8 +521,6 @@ private:
 };
 
 
-#endif /* ! USE_TBB_ATOMIC */
-
 
 #ifdef NOTHREADS
 
@@ -489,16 +541,9 @@ typedef atomic<long long> atomic_ll;
 typedef null_mutex spin_mutex;
 typedef null_lock<spin_mutex> spin_lock;
 
-#elif USE_TBB_SPINLOCK
-
-// Use TBB's spin locks
-typedef tbb::spin_mutex spin_mutex;
-typedef tbb::spin_mutex::scoped_lock spin_lock;
-
-
 #else
 
-// Define our own spin locks.  Do we trust them?
+// Define our own spin locks.
 
 
 /// A spin_mutex is semantically equivalent to a regular mutex, except
@@ -549,31 +594,41 @@ public:
         // found that OIIO_UNLIKELY makes this just a bit faster on 
         // gcc x86/x86_64 systems.
         while (! OIIO_UNLIKELY(try_lock())) {
-            do {
-                backoff();
-            } while (m_locked);
-
+#if OIIO_THREAD_ALLOW_DCLP
             // The full try_lock() involves a compare_and_swap, which
             // writes memory, and that will lock the bus.  But a normal
             // read of m_locked will let us spin until the value
             // changes, without locking the bus. So it's faster to
             // check in this manner until the mutex appears to be free.
+            // HOWEVER... Thread Sanitizer things this is an instance of
+            // an unsafe "double checked lock pattern" (DCLP) and flags it
+            // as an error. I think it's a false negative, because the
+            // outer loop is still an atomic check, the inner non-atomic
+            // loop only serves to delay, and can't lead to a true data
+            // race. But we provide this build-time switch to, at least,
+            // give a way to use tsan for other checks.
+            do {
+                backoff();
+            } while (m_locked);
+#else
+            backoff();
+#endif
         }
     }
 
     /// Release the lock that we hold.
     ///
     void unlock () {
-#if defined(__GNUC__) && (defined(__x86_64__) || defined(__i386__))
         // Fastest way to do it is with a store with "release" semantics
-        __asm__ __volatile__("": : :"memory");
-        m_locked = 0;
-        // N.B. GCC gives us an intrinsic that is even better, an atomic
-        // assignment of 0 with "release" barrier semantics:
-        //  __sync_lock_release (&m_locked);
-        // But empirically we found it not as performant as the above.
+#if defined(OIIO_USE_GCC_NEW_ATOMICS)
+        __atomic_clear (&m_locked, __ATOMIC_RELEASE);
+#elif defined(USE_GCC_ATOMICS)
+        __sync_lock_release (&m_locked);
+        //   Equivalent, x86 specific code:
+        //   __asm__ __volatile__("": : :"memory");
+        //   m_locked = 0;
 #elif defined(_MSC_VER)
-        _ReadWriteBarrier();
+        MemoryBarrier ();
         m_locked = 0;
 #else
         // Otherwise, just assign zero to the atomic (but that's a full 
@@ -585,10 +640,9 @@ public:
     /// Try to acquire the lock.  Return true if we have it, false if
     /// somebody else is holding the lock.
     bool try_lock () {
-#if USE_TBB_ATOMIC
-        // TBB's compare_and_swap returns the original value
-        return (*(atomic_int *)&m_locked).compare_and_swap (0, 1) == 0;
-#elif defined(__GNUC__)
+#if defined(OIIO_USE_GCC_NEW_ATOMICS)
+        return __atomic_test_and_set (&m_locked, __ATOMIC_ACQUIRE) == 0;
+#elif defined(USE_GCC_ATOMICS)
         // GCC gives us an intrinsic that is even better -- an atomic
         // exchange with "acquire" barrier semantics.
         return __sync_lock_test_and_set (&m_locked, 1) == 0;
@@ -605,14 +659,20 @@ public:
         lock_guard (spin_mutex &fm) : m_fm(fm) { m_fm.lock(); }
         ~lock_guard () { m_fm.unlock(); }
     private:
-        lock_guard(); // Do not implement (even though TBB does)
+        lock_guard(); // Do not implement
         lock_guard(const lock_guard& other); // Do not implement
         lock_guard& operator = (const lock_guard& other); // Do not implement
         spin_mutex & m_fm;
     };
 
 private:
+#if defined(OIIO_USE_GCC_NEW_ATOMICS)
+    // Using the gcc >= 4.8 new atomics, we can easily do a single byte flag
+    volatile char m_locked; ///< Atomic counter is zero if nobody holds the lock
+#else
+    // Otherwise, fall back on it being an int
     volatile int m_locked;  ///< Atomic counter is zero if nobody holds the lock
+#endif
 };
 
 
@@ -668,8 +728,13 @@ public:
         // Spin until the last reader is done, at which point we will be
         // the sole owners and nobody else (reader or writer) can acquire
         // the resource until we release it.
+#if OIIO_THREAD_ALLOW_DCLP
         while (*(volatile int *)&m_readers > 0)
                 ;
+#else
+        while (m_readers > 0)
+                ;
+#endif
     }
 
     /// Release the writer lock.
@@ -679,6 +744,18 @@ public:
         m_locked.unlock ();
     }
 
+    /// Acquire an exclusive ("writer") lock.
+    void lock () { write_lock(); }
+
+    /// Release an exclusive ("writer") lock.
+    void unlock () { write_unlock(); }
+
+    /// Acquire a shared ("reader") lock.
+    void lock_shared () { read_lock(); }
+
+    /// Release a shared ("reader") lock.
+    void unlock_shared () { read_unlock(); }
+
     /// Helper class: scoped read lock for a spin_rw_mutex -- grabs the
     /// read lock upon construction, releases the lock when it exits scope.
     class read_lock_guard {
